'''
Created on Jul 26, 2017

@author: runsheng
'''
import os
import json
import numpy as np
import pandas as pd
from scipy.spatial import distance
from collections import defaultdict

from keras.models import load_model
from sklearn.externals import joblib

from descriptors import descriptor_calculator

class qsar():
    def __init__(self, model_name):
        self.model = self._load_model(model_name)
        self.scalar = self._load_scalar(model_name)
        self.filter = self._load_filter(model_name)
        self.ad_dict = self._load_ad(model_name)
     
    def predict(self, SMILEs):
        SMILEs = [SMILEs] # temp solution...
        
        if len(SMILEs) > 1:
            raise ValueError("Only accept sinlge SMILEs") 
        
        descriptors = self.scalar.transform(descriptor_calculator.calculate(SMILEs, self.filter))
        inside_ad, error_bars = self._calculate_ad(descriptors, self.ad_dict)
        prediction = self.model.predict(descriptors) # in -log10(mol/L)
        prediction = self._unit_convertor(prediction) # in umol/L
        prediction_higher, prediction_lower = self._get_range(prediction[0], error_bars[0])
        
        return prediction[0][0], inside_ad[0], error_bars[0], prediction_higher[0], prediction_lower[0]
    
    def _calculate_ad(self, descriptors, ad_dict):
        cut_off_threshold = ad_dict['cut off']
        error_inside = ad_dict['error inside']
        error_outside = ad_dict['error outside']
        c = ad_dict['centroid']
        
        distance_to_c = [distance.euclidean(i, c) for i in descriptors]
        inside_ad = [1 if d < cut_off_threshold else 0 for d in distance_to_c]
        error_bars = [error_inside if k == 1 else error_outside for k in inside_ad]
        
        return inside_ad, error_bars
    
    def _unit_convertor(self, raw_pred):
        '''
        convert -log10(mol/L) to umol/L
        '''
        return [np.power(10, -raw_pred[0])*1e6]
        
        
    def _get_range(self, prediction, error_bars):
        return prediction + prediction*error_bars, prediction - prediction*error_bars
    
    def _load_model(self, model_name):
        cur_path = os.path.dirname(__file__)
        return load_model(cur_path+'/../models/'+model_name+'/model.h5')
    
    def _load_scalar(self,model_name):
        cur_path = os.path.dirname(__file__)
        return joblib.load(cur_path+'/../models/'+model_name+'/scaler.pkl')
    
    def _load_filter(self,model_name):
        cur_path = os.path.dirname(__file__)
        with open(cur_path+'/../models/'+model_name+'/filter.txt') as myfile:
            content = myfile.readlines()
        return [x.strip() for x in content]
    
    def _load_ad(self, model_name):
        cur_path = os.path.dirname(__file__)
        with open(cur_path+'/../models/'+model_name+'/ad.json') as ad_data:
            ad_dict = json.load(ad_data)
        
        return ad_dict
    
class run_all():
    @staticmethod
    def run(SMILEs, verbose=True):
        cur_path = os.path.dirname(__file__)
        all_models = [d for d in os.listdir(cur_path+'/../models') if os.path.isdir(os.path.join(cur_path+'/../models', d))]
        species = []
        all_p = defaultdict(list)
 
        for each_model in all_models:
            if verbose:
                print each_model
            species.append(each_model)
            this_qsar = qsar(each_model)
            this_p, this_inside, this_error, this_higher, this_lower = this_qsar.predict(SMILEs)
            all_p[each_model] = [this_p, this_inside, this_error, this_higher, this_lower]

        df = pd.DataFrame.from_dict(all_p, orient='index')
        df.columns = ['Prediction', 'Inside AD', 'Prediction Error', 'Prediction Upper', 'Prediction Lower']
        return df
    
if __name__ == '__main__':
    pass
#     # test
# #     model_name = "Oncorhynchus Mykiss"
# #        
# #     q = qsar(model_name=model_name)
# #     print q.predict(['CCC', 'C[N+](C)(C)CCCl.[Cl-]','ClC(Cl)(Cl)C(Cl)(Cl)Cl','c1ccccc1','ClC(Cl)(Cl)SN1C(=O)c2ccccc2C1=O','CCSC(=O)N(CC(C)C)CC(C)C'])
#     print run_all.run(['CCC'])
   
    


'''
Created on Jul 26, 2017

@author: runsheng
'''
import pandas as pd
from rdkit import Chem
from mordred import Calculator, descriptors

import warnings

class descriptor_calculator():
    def __init__(self):
        pass

    @staticmethod
    def calculate(SMILEs, filter=None):
        
        calc = Calculator(descriptors, ignore_3D=True)
        d = []
        for smi in SMILEs:
            try:
                m = Chem.MolFromSmiles(smi)
                d.append(calc(m))
            except:
                # The input SMILEs is invaild
                raise ValueError("Bad SMILEs Detected. Please Check: "+smi)
#                 warnings.warn("Bad SMILEs  Detect. Filling NA Values: "+smi)
#                 d.append(['NA'] * len(calc))
        
        d_df = pd.DataFrame(d, index=SMILEs, columns=[str(e_d) for e_d in calc.descriptors]).apply(pd.to_numeric, errors='coerce')
        if filter:
            d_df = d_df.loc[:,filter]
        
        d_df.fillna(0, inplace=True)
        return d_df.values

if __name__ == '__main__':
    # test
    pass


'''
Created on Jul 26, 2017

Pipeline to create Species Sensitivity Distribution Curves

@author: runsheng
'''

import pandas as pd
import numpy as np
from scipy.stats import lognorm
from scipy.stats import norm
import matplotlib.pyplot as plt



class ssd_generator():
    def __init__(self):
        pass
    
    def generate(self, 
                 ssd_df,
                 dist=lognorm, 
                 run_error =True,
                 run_bootstrap=True,
                 bootstrap_time=1000,
                 display_range=[0,100]):
        '''
        Input: pandas dataframe of Species_Name: LC50 Values
        Output: SSD Curves
        '''
        df_mean, frac_mean, df_high, frac_high, df_low, frac_low = self._frac(ssd_df, fraction=0.5) # dataframe is sorted here
        
        shape_mean, loc_mean, scale_mean = dist.fit(df_mean['Prediction'], floc=0)
        shape_lower, loc_lower, scale_lower = dist.fit(df_high['Prediction Lower'], floc=0)
        shape_upper, loc_upper, scale_upper = dist.fit(df_low['Prediction Upper'], floc=0)
        
        newx = pow(np.linspace(np.log10(0.01), np.log10(max(df_mean['Prediction'])), 1000), 10)
        
        fit_mean = dist.cdf(newx, s=shape_mean, loc=0, scale=scale_mean)
        fit_lower = dist.cdf(newx, s=shape_lower, loc=0, scale=scale_lower)
        fit_upper = dist.cdf(newx, s=shape_upper, loc=0, scale=scale_upper)
        
        # plot
        fig, ax = plt.subplots(figsize=(11,10))
        if run_bootstrap:
            this_boots = bootstrap()
            boots_res = this_boots.run_boots(df_mean, newx, shape_mean, scale_mean, dist=dist, times=bootstrap_time)
            upper, lower = this_boots.get_upper_lower(boots_res, upper=97.5, lower=2.5)
            for i in range(len(boots_res)):
                ax.plot(newx, boots_res[i], c='steelblue', alpha=0.2, zorder=0)
            # plot the upper and lower CI
            ax.plot(newx, upper,':', c='black', label='Upper and Lower Bound of Bootstrapping SSD')
            ax.plot(newx, lower,':', c='black')
            
        #plot the average curve
        ax.plot(newx, fit_mean,color='red',label='Average SSD')
        if run_error:
            ax.plot(newx, fit_lower,color='lightsalmon',label='Lower Bound')
            ax.plot(newx, fit_upper,color='lightsalmon',label='Upper Bound')
             
        #plot the points
        ax.scatter(df_mean['Prediction'], frac_mean, c='black', zorder=10)
        if run_error:
            ax.scatter(df_low['Prediction Lower'], frac_low, marker='x',c='gray', zorder=10)
            ax.scatter(df_high['Prediction Upper'], frac_high, marker='x', c='gray', zorder=10)
            
            
        # annotation
        for i, txt in enumerate(df_mean.index):
            ax.annotate(txt, (df_mean['Prediction'].values[i], frac_mean[i]))
        if run_error:
            for i, txt in enumerate(df_low.index):
                ax.annotate(txt, (df_low['Prediction Lower'].values[i]-0.15, frac_low[i]), size='small',color='gray')
                
            for i, txt in enumerate(df_high.index):    
                ax.annotate(txt, (df_high['Prediction Upper'].values[i], frac_high[i]-0.03), size='small',color='gray')
            
        # configeration
        plt.xscale('log')
        plt.xlim(display_range)
        plt.grid()
        plt.xlabel('Concentration (umol/L), in Log Scale')
        plt.ylabel('Cumulative Probability')
        plt.legend(loc='upper left')
            
        plt.show()
        return df_mean, df_high, df_low, fig
            
    def _frac(self, df, fraction=0.5):
        df_mean = df.sort_values(['Prediction'])
        frac_mean = self._ppoints(df['Prediction'], fraction)
        
        df_high = df.sort_values(['Prediction Upper'])
        frac_high = self._ppoints(df['Prediction Upper'], fraction)
        
        df_low = df.sort_values(['Prediction Lower'])
        frac_low = self._ppoints(df['Prediction Lower'], fraction)
        
        return df_mean, frac_mean, df_high, frac_high, df_low, frac_low
       
    def _ppoints(self, n, a):
        """ numpy analogue or `R`'s `ppoints` function
        see details at http://stat.ethz.ch/R-manual/R-patched/library/stats/html/ppoints.html 
        :param n: array type or number
        """
        try:
            n = np.float(len(n))
        except TypeError:
            n = np.float(n)
        return (np.arange(n) + 1 - a)/(n + 1 - 2*a)
    

class bootstrap():
    def __init__(self):
        pass
    
    def run_boots(self, df,newx, shape, scale, dist=lognorm, times=1000):
        res = []
        for i in range(times):
            this_boot = self._boots(df,newx, shape, scale, dist=dist)
            res.append(this_boot)
        return np.asarray(res)
    
    def get_upper_lower(self, boots_res, upper=97.5, lower=2.5):
        return (np.percentile(boots_res, upper, axis=0), (np.percentile(boots_res, lower, axis=0)))
        
    def _boots(self, df, newx, shape, scale, dist=lognorm):
        xr = lognorm.rvs(size=len(df['Prediction']), s=shape, loc=0, scale=scale)
        this_shape, this_loc, this_scale = lognorm.fit(xr, floc=0)
        this_fit = dist.cdf(newx, s=this_shape, loc=0, scale=this_scale)
        
        return list(this_fit)
    
    
        
if __name__ == '__main__':
    # test
    import io
    import requests
    
    url="https://raw.githubusercontent.com/EDiLD/r-ed/master/post_ssd/ssd_data.csv"
    s=requests.get(url).content
    c=pd.read_csv(io.StringIO(s.decode('utf-8')))
    c.set_index('species',inplace=True)
    
    this_generator = ssd_generator()
    this_generator.generate(c, dist=lognorm, run_bootstrap=True, bootstrap_time=1000, display=True)
    
    


from distutils.core import setup
from setuptools import find_packages

setup(
    name='QSAR_SSD_Toolbox',
    version='0.18',
    author="Runsheng Song",
    author_email="runsheng@umail.ucsb.edu",
    packages=find_packages(),
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    include_package_data=True,
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    install_requires=['mordred==0.3.1',
'scikit-learn==0.18.1',
'Keras==2.0.6',
'h5py==2.7.0',
'tensorflow==1.1.0'],
)



'''
Created on Aug 5, 2017

@author: runsheng
'''
import sys

from scipy.stats import lognorm

from src.qsar import *

if __name__ == '__main__':
    SMILEs = 'CCc1ccccc1' # The input SMILEs must be a list

#     this_q = qsar("Americamysis bahia")
#     print this_q.predict(SMILEs)
    this_q = run_all.run(SMILEs)
#     
#     print this_q.predict(SMILEs)
    
    from QSAR_SSD_Toolbox.src.ssd import ssd_generator

    this_ssd = ssd_generator()
    df_mean, df_high, df_low, this_fig = this_ssd.generate(this_q, dist=lognorm, run_error=True,run_bootstrap=False, bootstrap_time=500, display_range=[10,1000])
# 
#     print df_mean
#     print df_high
#     print df_low
    